/*
 authutil	access the authorization policy database (/etc/authorization)
 Copyright (C) 2005  Brian D. Wells
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 Please note: While not required by the license, you are invited to
 share your modifications with the current maintainer so that these may
 be incorporated into future releases of the software.
 
*/
 
#include <Security/Security.h>
#import <Foundation/Foundation.h>

#define writeStdOut(X...)	[(NSFileHandle *)[NSFileHandle fileHandleWithStandardOutput] writeData:[X dataUsingEncoding:NSASCIIStringEncoding]]
#define writeStdErr(X...)	[(NSFileHandle *)[NSFileHandle fileHandleWithStandardError] writeData:[X dataUsingEncoding:NSASCIIStringEncoding]]
#define writeStdOutWithFormat(X...)	writeStdOut([NSString stringWithFormat:X])
#define writeStdErrWithFormat(X...)	writeStdErr([NSString stringWithFormat:X])

NSDictionary *getRight(NSString *rightName);
OSStatus getRightQuietly(NSString *rightName, NSDictionary **definition);
OSStatus setRight(NSString *rightName, id definition);
OSStatus removeRight(NSString *rightName);
void unexpectedArgs(NSArray *arguments, unsigned offset);
id plistFromString(NSString *dataString);
id getValue(NSArray *arguments, int *idx, id existingValue);
NSString *nextArg(NSArray *arguments, int *offset);

AuthorizationRef auth;

int main (int argc, const char * argv[]) {
	NSArray *arguments;
	OSStatus err;
	NSDictionary *definition;
	BOOL invalidSyntax = NO;
	int result = 0;
	NSString *command;
	NSString *rightName;
	NSString *attribute;
	
    NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
	
	err = AuthorizationCreate(NULL, NULL, kAuthorizationFlagDefaults, &auth);
	if (err == errAuthorizationSuccess) {

		arguments = [[NSProcessInfo processInfo] arguments];
	
		if ([arguments count] > 2) {

			command = (NSString *)[arguments objectAtIndex:1];
			rightName = (NSString *)[arguments objectAtIndex:2];
			if ([command isEqualToString:@"read"]) {
				// get definition
				definition = getRight(rightName);
				if (definition != nil) {
					if ([arguments count] > 3) {
						// find attribute
						int idx = 3;
						id value = definition;
						do {
							attribute = nextArg(arguments, &idx);
							if (attribute != nil) {
								if ([value isKindOfClass:[NSDictionary class]]) {
									value = [(NSDictionary *)value objectForKey:attribute];
								} else if ([value isKindOfClass:[NSArray class]]) {
									int x;
									NSScanner *scanner = [NSScanner scannerWithString:attribute];
									if (![scanner scanInt:&x] || ![scanner isAtEnd]) {
										x = -1;
									}
									if (x >= 0 && x < [(NSArray *)value count]) {
										value = [(NSArray *)value objectAtIndex:x];
									} else {
										value = nil;
									}
								} else {
									value = nil;
								}
							}
						} while (attribute != nil && value != nil); 
						if (value == nil) {
							if ([arguments count] == 4) {
								attribute = [arguments objectAtIndex:3];
							} else {
								attribute = [NSString stringWithFormat:@"%@",[arguments subarrayWithRange:NSMakeRange(3,[arguments count] - 3)]];
							}
							writeStdErrWithFormat(@"The right/attribute pair of (%@, %@) does not exist\n",rightName,attribute);
						} else {
							writeStdOutWithFormat(@"%@\n",value);
						}
					} else {
						// show entire definition
						writeStdOutWithFormat(@"%@\n",definition);
					}
					[definition autorelease];
				}				
			} else if ([command isEqualToString:@"read-type"]) {
				if ([arguments count] < 4) {
					invalidSyntax = YES;
				} else {
					definition = getRight(rightName);
					if (definition != nil) {
						// find attribute
						int idx = 3;
						id value = definition;
						do {
							attribute = nextArg(arguments, &idx);
							if (attribute != nil) {
								if ([value isKindOfClass:[NSDictionary class]]) {
									value = [(NSDictionary *)value objectForKey:attribute];
								} else if ([value isKindOfClass:[NSArray class]]) {
									int x;
									NSScanner *scanner = [NSScanner scannerWithString:attribute];
									if (![scanner scanInt:&x] || ![scanner isAtEnd]) {
										x = -1;
									}
									if (x >= 0 && x < [(NSArray *)value count]) {
										value = [(NSArray *)value objectAtIndex:x];
									} else {
										value = nil;
									}
								} else {
									value = nil;
								}
							}
						} while (attribute != nil && value != nil); 
						if (value == nil) {
							if ([arguments count] == 4) {
								attribute = [arguments objectAtIndex:3];
							} else {
								attribute = [NSString stringWithFormat:@"%@",[arguments subarrayWithRange:NSMakeRange(3,[arguments count] - 3)]];
							}
							writeStdErrWithFormat(@"The right/attribute pair of (%@, %@) does not exist\n",rightName,attribute);
						} else {
							writeStdOut(@"Type is ");
							if ([value isKindOfClass:[NSString class]]) {
								writeStdOut(@"string");
							} else if ([value isKindOfClass:[NSData class]]) {
								writeStdOut(@"data");
							} else if ([value isKindOfClass:[NSDate class]]) {
								writeStdOut(@"date");
							} else if ([value isKindOfClass:[NSDictionary class]]) {
								writeStdOut(@"dictionary");
							} else if ([value isKindOfClass:[NSArray class]]) {
								writeStdOut(@"array");
							} else if ([value isKindOfClass:[NSNumber class]]) {
								NSString *objCtypeStr = [NSString stringWithUTF8String:[(NSNumber *)value objCType]];
								if ([objCtypeStr rangeOfCharacterFromSet:[NSCharacterSet characterSetWithCharactersInString:@"islqISLQ"]].location != NSNotFound) {
									writeStdOut(@"integer");
								} else if ([objCtypeStr rangeOfCharacterFromSet:[NSCharacterSet characterSetWithCharactersInString:@"cCB"]].location != NSNotFound) {
									writeStdOut(@"boolean");
								} else if ([objCtypeStr rangeOfCharacterFromSet:[NSCharacterSet characterSetWithCharactersInString:@"fd"]].location != NSNotFound) {
									writeStdOut(@"float");
								} else {
									writeStdOutWithFormat(@"unknown number type '%@'",objCtypeStr);
								}
							} else {
								writeStdOut(@"unknown");
							}
							writeStdOut(@"\n");	
						}
						[definition autorelease];
					}
				}

			} else if ([command isEqualToString:@"write"] || [command isEqualToString:@"write-test"]) {
				if ([arguments count] < 4) {
					// not enough arguments
					invalidSyntax = YES;
				} else {
					if ([arguments count] == 4) {
						// argument 3 must be definition to overwrite
						id plist = plistFromString((NSString *)[arguments objectAtIndex:3]);
						if (plist != nil) {
							// got definition...
							// make sure it is NSString or NSDictionary
							if ([plist isKindOfClass:[NSString class]] || [plist isKindOfClass:[NSDictionary class]]) {
								if ([command isEqualToString:@"write-test"]) {
									if ([plist isKindOfClass:[NSString class]]) {
										writeStdOutWithFormat(@"Test Mode: would write out right '%@' with delegate rule '%@'\n\n",rightName,plist);
									} else {
										writeStdOutWithFormat(@"Test Mode: would write out right '%@' with definition:\n%@\n\n",rightName,plist);
									}
								} else {
									err = setRight(rightName, plist);
								}
							} else {
								writeStdErr(@"property list or name of write to duplicate not specified\n");
							}
						}
					} else {
						// 5 or more arguments means a list of argument, value
						err = getRightQuietly(rightName, &definition);	// always an NSDictionary or nil

						int idx = 0;
						NSMutableArray *newArguments = [NSMutableArray arrayWithCapacity:10];
						[newArguments addObject:@"-dict-add"];
						[newArguments addObjectsFromArray:[arguments subarrayWithRange:NSMakeRange(3,[arguments count] - 3)]];
						
						definition = (NSDictionary *)getValue(newArguments, &idx, definition);	// should be NSDictionary or nil
						if (definition == nil) {
							writeStdErr(@"an error occured with arguments\n");
						} else {
							if ([command isEqualToString:@"write-test"]) {
								writeStdOutWithFormat(@"Test Mode: would write out right '%@' with definition:\n%@\n\n",rightName,definition);
							} else {
								// set right if definition is not nil
								err = setRight(rightName, definition);
							}
						}
						
					}
				}
			} else if ([command isEqualToString:@"delete"] || [command isEqualToString:@"delete-test"]) {
				// delete
				if ([arguments count] == 3) {
					if ([command isEqualToString:@"delete-test"]) {
						err = getRightQuietly(rightName,&definition);
						if (definition == nil) {
							writeStdOutWithFormat(@"Test Mode: would fail since right '%@' does not exist\n\n",rightName);
						} else {
							writeStdOutWithFormat(@"Test Mode: would remove right '%@'\n\n",rightName);
						}
					} else {
						removeRight(rightName);
					}
				} else {
					definition = getRight(rightName);
					if (definition != nil) {
						// find attribute to delete
						int idx = 3;
						NSMutableDictionary *newDefinition = [[definition mutableCopy] autorelease];
						id parent = newDefinition;
						NSString *valueName = nil;
						id value = nil;
						do {
							attribute = nextArg(arguments,&idx);
							if (attribute != nil) {
								if (value != nil) {
									// make mutable copy of value
									if ([value conformsToProtocol:@protocol(NSMutableCopying)]) {
										value = [[value mutableCopy] autorelease];
										if ([parent isKindOfClass:[NSMutableDictionary class]]) {
											[(NSMutableDictionary *)parent setObject:value forKey:valueName];
										} else if ([parent isKindOfClass:[NSMutableArray class]]) {
											[(NSMutableArray *)parent replaceObjectAtIndex:[valueName intValue] withObject:value];
										}
										// should not be any other kind of value
									}
									parent = value;
								}
								valueName = attribute;
								if ([parent isKindOfClass:[NSDictionary class]]) {
									value = [(NSDictionary *)parent objectForKey:attribute];
								} else if ([parent isKindOfClass:[NSArray class]]) {
									int x;
									NSScanner *scanner = [NSScanner scannerWithString:attribute];
									if (![scanner scanInt:&x] || ![scanner isAtEnd]) {
										x = -1;
									}
									if (x >= 0 && x < [(NSArray *)parent count]) {
										value = [(NSArray *)parent objectAtIndex:x];
									} else {
										value = nil;
									}
								} else {
									value = nil;
								}
							}
						} while (attribute != nil && value != nil); 
						if (value == nil) {
							if ([arguments count] == 4) {
								attribute = [arguments objectAtIndex:3];
							} else {
								attribute = [NSString stringWithFormat:@"%@",[arguments subarrayWithRange:NSMakeRange(3,[arguments count] - 3)]];
							}
							writeStdErrWithFormat(@"The right/attribute pair of (%@, %@) does not exist\n",rightName,attribute);
						} else {
							// remove value from parent
							if ([parent isKindOfClass:[NSMutableDictionary class]]) {
								[(NSMutableDictionary *)parent removeObjectForKey:valueName];
							} else if ([parent isKindOfClass:[NSMutableArray class]]) {
								[(NSMutableArray *)parent removeObjectAtIndex:[valueName intValue]];
							}
							// should not be any other kind of value
							if ([command isEqualToString:@"delete-test"]) {
								writeStdOutWithFormat(@"Test Mode: would write out right '%@' with definition:\n%@\n\n",rightName,newDefinition);
							} else {
								err = setRight(rightName, newDefinition);
							}
						}
					}					
				}
			} else {
				invalidSyntax = YES;
			}
		} else {
			invalidSyntax = YES;
		}
	
		if (invalidSyntax) {
			writeStdErr(@"Command line interface to authorization policy database (/etc/authorization).\nSyntax:\n\n");
			writeStdErrWithFormat(@"'%@' followed by one of the following:\n\n",[[NSProcessInfo processInfo] processName]);
			writeStdErr(@"  read <right>                       shows definition for given right\n");
			writeStdErr(@"  read <right> <attribute>           shows definition for given right, attribute\n\n");
			writeStdErr(@"  read-type <right> <attribute>      shows the type for given right, attribute\n\n");
			writeStdErr(@"  write <right> <definition>         writes right (overwrites existing)\n");
			writeStdErr(@"  write <right> <attribute> <value>  write attribute for given right\n\n");
			writeStdErr(@"  delete <right>                     deletes specified right\n");
			writeStdErr(@"  delete <right> <attribute>         deletes attribute for given right\n\n");
			writeStdErr(@"  help                               print this help\n\n");
			writeStdErr(@"<definition> is one of\n");
			writeStdErr(@"  text representation of right definition in property list format\n");
			writeStdErr(@"  name of a delegate rule (authenticate-admin, authenticate-session-owner, etc)\n\n");
			writeStdErr(@"<value> is one of:\n");
			writeStdErr(@"  <value_rep>\n  -string <string_value>\n  -data <hex_digits>\n  -int[eger] <integer_value>\n");
			writeStdErr(@"  -float  <floating-point_value>\n  -bool[ean] (true | false | yes | no)\n  -date <date_rep>\n");
			writeStdErr(@"  -array <value1> <value2> ...\n  -array-add <value1> <value2> ...\n");
			writeStdErr(@"  -dict <key1> <value1> <key2> <value2> ...\n  -dict-add <key1> <value1> ...\n\n");
		}
		AuthorizationFree(auth, kAuthorizationFlagDestroyRights);
	}
	[pool release];
    return result;
}

NSDictionary *getRight(NSString *rightName) {
	OSStatus err;
	NSDictionary *definition;

	err = getRightQuietly(rightName,&definition);
	if (err == errAuthorizationSuccess) {
		return definition;
	} else if (err == errAuthorizationDenied) {
		writeStdErrWithFormat(@"Right %@ does not exist\n",rightName);
	} else {
		writeStdErrWithFormat(@"\nAuthorizationRightGet error: %d\n\n",err);
	}
	return nil;
}

OSStatus getRightQuietly(NSString *rightName, NSDictionary **definition) {
	OSStatus err;
	
	err = AuthorizationRightGet([rightName UTF8String],(CFDictionaryRef *)definition);
	if (err != errAuthorizationSuccess) {
		*definition = nil;
	}
	return err;
}

OSStatus setRight(NSString *rightName, id definition) {
	OSStatus err;

	if ([definition isKindOfClass:[NSString class]] || [definition isKindOfClass:[NSDictionary class]]) {
		err = AuthorizationRightSet(auth, [rightName UTF8String], definition, NULL, NULL, NULL);
		if (err != errAuthorizationSuccess) {
			writeStdErrWithFormat(@"\nAuthorizationRightSet error: %d\n\n",err);
		}
					
	} else {
		writeStdErrWithFormat(@"Unsupported property list type: %@\n", [definition className]);
	}
	return err;
}

OSStatus removeRight(NSString *rightName) {
	OSStatus err;
	
	err = AuthorizationRightRemove(auth, [rightName UTF8String]);
	if (err != errAuthorizationSuccess) {
		writeStdErrWithFormat(@"\nAuthorizationRightRemove error: %d\n\n",err);
	}
	return err;
}

void unexpectedArgs(NSArray *arguments, unsigned offset) {
	// unexpected arguments
	NSRange range;
	range.location = offset;
	range.length = [arguments count] - offset;
	writeStdErrWithFormat(@"Unexpected argument%s: ",(range.length == 1 ? "": "s"));
	writeStdErrWithFormat(@"%@\n",[arguments subarrayWithRange:range]);
}

NSString *nextArg(NSArray *arguments, int *offset) {
	NSString *result = nil;
	if (*offset < [arguments count]) {
		result = [arguments objectAtIndex:*offset];
		(*offset)++;
	}
	return result;
}

id plistFromString(NSString *dataString) {
	id plist;
NS_DURING
	plist = [dataString propertyList];
NS_HANDLER
 	plist = nil;
NS_ENDHANDLER
	return plist;
}

id getValue(NSArray *arguments, int *idx, id existingValue) {
	NSString *argument;
	id result = nil;

	// first argument is value or option
	argument = nextArg(arguments, idx);
	if (argument != nil) {
		if ([argument isEqualToString:@"-string"]) {
			// NSString
			result = nextArg(arguments, idx);
		} else if ([argument isEqualToString:@"-data"]) {
			// NSData
			 result = plistFromString([NSString stringWithFormat:@"<%@>",nextArg(arguments,idx)]);
 		} else if ([argument isEqualToString:@"-int"] || [argument isEqualToString:@"-integer"]) {
			// NSNumber (integer)
			argument = nextArg(arguments,idx);
			if (argument != nil) {
				result = [NSNumber numberWithInt:[argument intValue]];
			}
		} else if ([argument isEqualToString:@"-float"]) {
			// NSNumber (float)
			argument = nextArg(arguments,idx);
			if (argument != nil) {
				result = [NSNumber numberWithFloat:[argument floatValue]];
			}
 		} else if ([argument isEqualToString:@"-bool"] || [argument isEqualToString:@"-boolean"]) {
			// NSNumber (boolean)
			argument = nextArg(arguments,idx);
			if (argument != nil) {
				if ([argument caseInsensitiveCompare:@"true"] == NSOrderedSame || [argument caseInsensitiveCompare:@"yes"] == NSOrderedSame) {
					result = [NSNumber numberWithBool:YES];
				} else if ([argument caseInsensitiveCompare:@"false"] == NSOrderedSame || [argument caseInsensitiveCompare:@"no"] == NSOrderedSame) {
					result = [NSNumber numberWithBool:NO];
				}
			}
		} else if ([argument isEqualToString:@"-date"]) {
			argument = nextArg(arguments,idx);
			if (argument != nil) {
				result = [NSDate dateWithString:argument];
				if (result == nil) {
					result = [NSDate dateWithNaturalLanguageString:argument];
				}
			}
		} else if ([argument isEqualToString:@"-array"] || [argument isEqualToString:@"-array-add"]) {
			// NSArray
			NSMutableArray *array = [NSMutableArray arrayWithCapacity:10];
			id value = nil;

			if (existingValue != nil && [argument isEqualToString:@"-array-add"] && [existingValue isKindOfClass:[NSArray class]]) {
				// existingValue is NSArray - copy over objects
				[array addObjectsFromArray:(NSArray *)existingValue]; 
			}
			
			do {
				value = getValue(arguments, idx, nil);
				if (value != nil) {
					[array addObject:value];
				}
			} while (value != nil && *idx < [arguments count]);
			if (value != nil) {
				result = [array copy];
			}
		} else if ([argument isEqualToString:@"-dict"] || [argument isEqualToString:@"-dict-add"]) {
			// NSDictionary
			NSMutableDictionary *dict = [NSMutableDictionary dictionaryWithCapacity:10];
			id value;

			if (existingValue != nil) {
				// existingValue should be NSDictionary
				if ([existingValue isKindOfClass:[NSDictionary class]] && [argument isEqualToString:@"-dict-add"]) {
					// copy over entries
					[dict addEntriesFromDictionary:(NSDictionary *)existingValue];
				} else {
					existingValue = nil;		// replace instead
				}
			}

			do {
				value = nil;
				argument = nextArg(arguments,idx);
				if (argument != nil) {

					if (existingValue != nil) {
						value = [dict objectForKey:argument];
					}
					
					value = getValue(arguments, idx, value);
					if (value != nil) {
						[dict setObject:value forKey:argument];
					}
				}
			} while (value != nil && *idx < [arguments count]);
			if (value != nil) {
				result = [dict copy];
			}
		} else {
			// guess
			NSScanner *scanner = [NSScanner scannerWithString:argument];
			// try integer
			int i;
			if ([scanner scanInt:&i] && [scanner isAtEnd]) {
				result = [NSNumber numberWithInt:i];
			}
			if (result == nil) {
				// try float
				float f;
				[scanner setScanLocation:0];
				if ([scanner scanFloat:&f] && [scanner isAtEnd]) {
					result = [NSNumber numberWithFloat:f];
				}
			}
			if (result == nil) {
				// try boolean
				if ([argument caseInsensitiveCompare:@"true"] == NSOrderedSame || [argument caseInsensitiveCompare:@"yes"] == NSOrderedSame) {
					result = [NSNumber numberWithBool:YES];
				} else if ([argument caseInsensitiveCompare:@"false"] == NSOrderedSame || [argument caseInsensitiveCompare:@"no"] == NSOrderedSame) {
					result = [NSNumber numberWithBool:NO];
				}
			}
			if (result == nil) {
				result = plistFromString(argument);
			}
			if (result == nil) {
				result = argument;
			}
		}
	}
	return result;
}
